package geniusweb.protocol.session.mopac.phase;

import java.util.List;
import java.util.stream.Collectors;

import com.fasterxml.jackson.annotation.JsonIgnore;

import geniusweb.actions.Action;
import geniusweb.actions.PartyId;
import geniusweb.actions.Votes;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.session.mopac.PartyStates;
import geniusweb.voting.VotingEvaluator;

public abstract class DefaultPhase implements Phase {
	// deadline for this phase, ms since 1970
	protected final Long deadline;
	protected final PartyStates partyStates;

	// don't serialize this, users don't need it..
	@JsonIgnore
	protected final VotingEvaluator evaluator;

	// don't serialize this, it will cause very large file
	@JsonIgnore
	protected final Phase prevPhase;

	/**
	 * 
	 * @param actions     the actions done in this phase .
	 * @param prevPhase   the previous phase, can be used to collect previous
	 *                    votes etc.
	 * @param partyStates
	 * @param deadline    deadline for this phase, ms since 1970
	 */
	public DefaultPhase(Phase prevPhase, PartyStates partyStates, Long deadline,
			VotingEvaluator evaluator) {
		this.partyStates = partyStates;
		this.deadline = deadline;
		this.prevPhase = prevPhase;
		this.evaluator = evaluator;
	}

	@Override
	public boolean isFinal(long now) {
		return now >= deadline || partyStates.getNotYetActed().isEmpty();
	}

	@Override
	public long getDeadline() {
		return deadline;
	}

	/**
	 * @return all votes done in this phase.
	 */
	public List<Votes> getVotes() {
		return partyStates.getActions().stream()
				.filter(act -> act instanceof Votes).map(act -> (Votes) act)
				.collect(Collectors.toList());
	}

	/**
	 * Check if actor can do given action. Basic checks:
	 * <ul>
	 * <li>real actor does not match action's actor
	 * <li>deadline for this phase has passed
	 * <li>action is not allowed in this phase
	 * <li>actor already acted in this phase
	 * </ul>
	 * 
	 * @param activeParties the currently active parties
	 * @param actor
	 * @param action
	 * @param timems
	 * @throws ProtocolException
	 * @throw ProtocolException if the action violates the protocol, or null if
	 *        ok
	 */
	protected void checkAction(PartyId actor, Action action, long timems)
			throws ProtocolException {
		if (action == null)
			throw new ProtocolException("Action=null", actor);
		if (!(actor.equals(action.getActor())))
			throw new ProtocolException(
					"Incorrect actor info in action:" + action.getActor(),
					actor);
		if (isFinal(timems))
			throw new ProtocolException("passed deadline", actor);
		if (!(getAllowedActions().contains(action.getClass())))
			throw new ProtocolException(
					"Action not allowed in " + this.getClass() + ":" + action,
					actor);
		if (!partyStates.getNotYetActed().contains(actor))
			throw new ProtocolException("Actor can not act anymore", actor);
		return;
	}

	/**
	 * @param timems current time
	 * @return current PartyStates
	 */
	@Override
	public PartyStates getPartyStates() {
		return partyStates;
	}

	@Override
	public VotingEvaluator getEvaluator() {
		return evaluator;
	}

	@Override
	public Phase next(long now, long duration) {
		if (duration < PHASE_MINTIME || duration > PHASE_MAXTIME)
			throw new IllegalArgumentException("Bug, illegal duration");
		if (!isFinal(now))
			throw new IllegalStateException(
					"next must only be called when phase isFinal");
		if (!partyStates.getNotYetActed().isEmpty())
			throw new IllegalStateException("Bug. State has not been finished");
		return checkedNext(now + duration);
	}

	/**
	 * 
	 * @param deadline
	 * @return the next phase
	 */
	protected abstract Phase checkedNext(long deadline);

	@Override
	public String toString() {
		// don't print prev state, it will give HUGE list because it also
		// contains prevstate.....
		return getClass().getSimpleName() + "[" + partyStates + "," + deadline
				+ "," + evaluator + "]";
	}
}
